/**
 ** This file is part of Webscan 
 **	a web site fuzzer for common vulnerabilities
 **
 ** Copyright (C) 2003 gunzip
 **
 ** 	<techieone@softhome.net>
 ** 	<avatar13@iname.com>
 **
 **	http://members.xoom.it/gunzip/
 **
 ** This program is free software; you can redistribute it and/or
 ** modify it under the terms of the GNU General Public License
 ** as published by the Free Software Foundation; either version 2
 ** of the License, or (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program; if not, write to the Free Software
 ** Foundation, Inc., 59 Temple Place - Suite 330,Boston, MA 02111-1307, USA.
 **
 **/

#include	"util.h"
#include	"webscan.h"

/**
 **	FIXME: cookies
 **/
static char * create_client_header( char * host, size_t datalen )
{
#	define	CRLF		"\r\n"
#	define	HHOST		"Host: "
#	define	HAGENT	"User-Agent: Links (2.1pre9; Linux 2.4.20 i686; 80x30)"
#	define	HACCEPT	"Accept: */*"
#	define	HCONN		"Connection: Close"
#	define	HCHAR		"Accept-Charset: iso-8859-1, utf-8;q=0.5, *;q=0.5"
#	define	HLANG		"Accept-Language: it, en"
#	define	HCONTLEN	"Content-Length: "
#	define	HCTYPE	"Content-Type: application/x-www-form-urlencoded"

	char	* chead = NULL;

	int size = 	strlen(HHOST) + strlen(HAGENT) + strlen(host) + 
					strlen(HACCEPT) + strlen(HCONN) + strlen(HCHAR) + 
					strlen(HLANG) + 32;

	if ( !host )
		return NULL;

	if ( !datalen )
	{
		chead = (char * )xmalloc( size );
	
		sprintf( chead, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s", 
					HHOST, host, CRLF,
					HAGENT,	CRLF,
					HACCEPT, CRLF,
					HCONN,	CRLF,
					HCHAR,	CRLF,
					HLANG,	CRLF,
					CRLF );
	}
	else
	{
		chead = (char * )xmalloc( size + strlen(HCONTLEN) + strlen(HCTYPE) );
	
		sprintf( chead, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%d%s%s", 
					HHOST, host, CRLF,
					HAGENT,	CRLF,
					HACCEPT, CRLF,
					HCONN,	CRLF,
					HCHAR,	CRLF,
					HLANG,	CRLF,
					HCTYPE,	CRLF,
					HCONTLEN, datalen,
					CRLF, CRLF );
	}
	return chead;
}

static char * xencode( char * url ) 
{
	char 	* ret		= NULL;
	char	* start 	= NULL;

	if ( ISNULL( url ) )
		return NULL;

	start = ret = (char *)xmalloc( strlen(url) * 3 + 1 );
	
	while(  !ISNULL(url) )
	{
		switch( *url )
		{
			case '/':
			case '?':
			case ':':
			case '@':
			case '=':
			case '+':
			case ';':
			case '&':
				*ret++ = *url;
				break;
			case '%':
				*ret++ = *url;
				if ( !ISNULL(url + 1) && !ISNULL(url + 2) 
						&& *(url + 1)== 0x30 && *(url + 2) == 0x30 )  {
					*ret++ = *(url + 1);
					*ret++ = *(url + 2);
					url += 2;  
				}
				break;
			default:
				sprintf( ret, "%%%02x", *url );
				ret += 3;
			}
			url++;
	}
	*ret = 0;
	ret = xstrdup( start );
	xfree( start );
	debug( "Line encoded = %s\n", ret );
	return ret;
}

static void get_status( char * resp, unsigned int * status )
{
	char	* ptr = NULL;

	if ( !resp )
		return;

	if ( (ptr = resp + strlen( "HTTP/1.x ")) )
	{
		switch ( atoi(ptr) )
		{
			case	200:	*status = OK;			break;
			case	301:	*status = MOVED;		break;
			case	302:	*status = FOUND;		break;
			case	400:	*status = BADREQ;		break;
			case	401:	*status = UNAUTH;		break;
			case	403:	*status = FORBID;		break;
			case	404:	*status = NOTFOUND;	break;
			case	405:	*status = NOTALLOW;	break;
			case	414:	*status = TOOLONG;	break;
			case	500:	*status = INTERR;		break;
			default:		*status = OK;			break;
		}
	}
}

/**
 ** DESCRIPTION
 ** 	connects to the httpd 
 **	then send the get request for specified page
 **
 ** RETURN VALUE
 **	a pointer to the stream and status information in &status
 **/
static FILE * http_request_r( char * host, 
										char * data, 
										char * page, 
										unsigned short port, 
										unsigned long ip, 
										int * recursion, 
										unsigned int * status,
										boolean encode )
{
	int	fd				= 0;
	char	* sa			= NULL;
	char	* ch 			= NULL;
	char	* header 	= NULL;
	char	* vpage 		= NULL;
	char	* vhost 		= NULL;
	char	* location 	= NULL;
	char	* encoded	= NULL;
	FILE	* stream 	= NULL;

	short	redir = 0;
	unsigned short vport = 80;
	int sock = connect_to_host( ip, port );

	if ( sock == -1 || !page || (*recursion)++ == MAXRECURSION )
		return NULL;

	if ( data )
		fprintf( stdout, " -- (POST) http://%s%s%s?%s [%c] ",	
					host, page[0] == '/' ? "" : "/", page, 
					data, encode ? 'E' : 'N' );
	else
		fprintf( stdout, " -- (GET) http://%s%s%s [%c] ",
				host, page[0] == '/' ? "" : "/", page,
				encode ? 'E' : 'N' );

	if ( (stream = fdopen( sock, "a+" )) == NULL )
	{
		ERROR( "Can't create stream" );
		return NULL;
	}
   if ( (fd = fileno( stream )) == -1 )
	{
		ERROR( "Can't get file descriptor" );
      return NULL;
	}

	if ( encode == TRUE ) encoded = xencode( page );

	if ( !net_fprintf( stream, "%s%s%s HTTP/1.0\r\n", 
			data ? "POST" : "GET",
			page[0] == '/' ? " " : " /", encoded ? encoded : page ) )
	{
		ERROR( "Can't send query" );
		return NULL;
	}

	if ( encode == TRUE ) xfree( encoded );

	if ( !net_fprintf( stream, "%s", 
			ch = create_client_header( host, data ? strlen(data) : 0 )) )
	{
		xfree( ch );
		ERROR( "Can't send client header" );
		return NULL;
	}
	if ( data )
	{
		if  ( !net_fprintf( stream, "%s\r\n", data ) )
		{
			ERROR( "Can't post data" );
			return NULL;
		}
	}	
	if ( !(header = get_header( stream )) )
	{
		ERROR( "Can't get header from server" );
		return NULL;
	}
	if ( (sa = get_server_answer( header )) )
	{
		if ( status ) get_status( sa, status );
		fprintf( stdout, KYA "(" DEF "%s" KYA ")\n" DEF, sa );
	}
	if ( (location = redirect_get_location( header )) )
	{
		vpage = url_get_link( location );
		vhost = url_get_host( location );
		vport = url_get_port( location ); 
			
		if ( !vhost )	vhost = host;

		if ( vpage )
		{
			if ( ISRELATIVE( vpage ) )
			{
				char tmp[ BUFSIZ ]; 
				char * dir = GETDIR(page);
				if ( dir )
				{
					memset( tmp, 0, sizeof(tmp) );
					snprintf( tmp, sizeof( tmp ), "%s%s", dir, vpage ); 
					vpage = tmp ;
					debug( "Redirecting to %s\n", vpage );
					xfree( dir );
				}
			}
			else
			{
				debug( "Redirecting to %s\n", vpage );
			}
			redir = 1;
		}
		else
		{
			ERROR( "Redirect failed\n" );
			redir = -1;
		}
	}

	xfree( ch );
	xfree( sa );
	xfree( header );
	xfree( location );

	if ( redir == 1 )
	{
		XCLOSE( fd, stream );
		if ( (ip = resolve( vhost )) != -1 )
			return( http_request_r( vhost, data, vpage, vport, 
								ip, recursion, status, encode ));
	}

	xfree( vhost );
	xfree( vpage );

	if ( redir != -1 ) 
		return stream;
	else return NULL;
}

/**
 **	FIXME: better using a structure pointer
 **/

FILE * http_request( char * host, char * data, char * page,
							unsigned short port, unsigned long ip,
							unsigned int * status, boolean encode )
{
	int first = 0;
	return http_request_r( host, data, page, port, ip, &first, status, encode );
}

char * http_head_request( unsigned long ip, unsigned short port )
{
	int sockfd	= 0;
	FILE * s 	= NULL;

	if ( (sockfd = connect_to_host( ip, port )) != -1 )
	{
		if ( (s = fdopen( sockfd, "a+" )) == NULL )
		{
			ERROR( "Can't create stream" );
			return NULL;
		}
		if ( net_fprintf( s, "HEAD / HTTP/1.0\r\n\r\n" ) != -1 )
			return get_header( s );
	}
	XCLOSE( sockfd, s );
	return NULL;
}
