/**
 ** This file is part of Webscan 
 **	a web site fuzzer for common vulnerabilities
 **
 ** Copyright (C) 2003 gunzip 
 **
 **   <techieone@softhome.net>
 **   <avatar13@iname.com>
 **
 **   http://members.xoom.it/gunzip/
 **
 ** This program is free software; you can redistribute it and/or
 ** modify it under the terms of the GNU General Public License
 ** as published by the Free Software Foundation; either version 2
 ** of the License, or (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program; if not, write to the Free Software
 ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 **
 **/

#include "util.h"
#include "webscan.h"

#include <time.h>
#include <signal.h>

/** LAME GLOBALS **/

char * __logfile 			= NULL ;
unsigned int __level 	= 0;

/** END GLOBALS **/

void sigexit_handler( int sig )
{
	fprintf( stdout, DEF );
	fprintf( stderr, DEF );
	signal( SIGINT, SIG_DFL );
	raise( SIGINT );
}

void usage( char * arg )
{
	fprintf( stdout, 	"\n\n__Usage: %s <url> [-o outputfile] [options]\n\n"
							"\t-s\tchecks for sql, asp, vb, php errors (default)\n"
							"\t-d\tchecks for directory traversal (*)\n"
							"\t-p\tchecks for insecure perl open or xss (*)\n"
							"\t-e\tchecks for execution through shell escapes or xss (*)\n"
							"\t-a\tsets all of the above switches on (*)\n\n"
							"(*) Very noisy scans (!)\n\n", arg );
}

int main(int argc, char *argv[])
{
	time_t			now;
	int				fd				= 0;
	int				ch				= 0;
	int 				count			= 0;
	unsigned int	status		= 0;
	boolean			allvisited 	= FALSE;
	boolean			refresh		= FALSE;
	struct node 	* htnode		= NULL;
	char				* host 		= NULL;
	char 				* page 		= NULL;
	char				* header 	= NULL;
	char				* newurl		= NULL;
	char				* dir 		= NULL;
	FILE 				* stream		= NULL;

	char           buf[ BUFSIZ ];

   unsigned long  ip          = 0;
   unsigned short port        = 80;

	time( &now );

	signal( SIGINT, sigexit_handler );
	
	fprintf( stdout, "%s %s\n", BANNER, VERSION );

	if ( argc < 2 )
	{
		usage( argv[0] );
		return -1;
	}

	if ( !(host = url_get_host( argv[1] )) )
	{
		host = xstrdup( argv[1] );
	}
	else
	{
		port = url_get_port( argv[1] );
		page = url_get_link( argv[1] );
		dir = GETDIR( page );
	}
	while ( (ch = getopt(argc, argv,"o:sdpea")) !=EOF )
	{
		switch( ch )
		{
			case 'o': __logfile = xstrdup( optarg ); break;
			case 's': __level |= TFSAP; break;
			case 'd': __level |= TFDIR; break;
			case 'p': __level |= TFPERL; break;
			case 'e': __level |= TFSHELL; break;	
			case 'a': __level = ( TFSAP | TFDIR | TFPERL | TFSHELL ); break;
			default:	usage(argv[0]); return -1;
		}
	}
	if ( !__logfile ) { 
			__logfile = (char *)xmalloc( strlen(host) + 8 ); 
			sprintf( __logfile, "log.%s", host );
	}
	if ( !__level ) __level |= TFSAP;

	if ( !page )
		page = xstrdup( DEFAULTPAGE );

	if ( !dir )
		dir = GETDIR( DEFAULTPAGE );

	for ( count = HT_SIZE - 1; count; count-- )
	{
		HTlinks[count] = NULL;
		HTforms[count] = NULL;
		HTtools[count] = NULL;
		HTusers[count] = NULL;
	}

	fflush( stdout );

	if ( (ip = resolve( host )) == -1 )
	{
		ERROR( "Can't resolve host" );
		exit( EXIT_FAILURE );
	}

	memset( buf, 0, sizeof buf );

	snprintf( buf, sizeof buf - 2,
				"\n-------------------------------------------------------------------------\n" 
				" WebScan of %s:%d [%s] (%s)\n %s" 
				"-------------------------------------------------------------------------\n\n", 
				host, port, page, ip_to_ascii( ip ), ctime( &now ));

	fprintf( stdout, buf );
	freport( __logfile, buf );

	if ( (header = http_head_request( ip, 80 )) == NULL )
	{
		ERROR( "Can't get server version" );
	}
	else
	{
		fprintf( stdout, RED "Server header:\n\n" DEF "%s\n\n", header );
		freport( __logfile, "Server header:\n\n%s\n\n", header );
		xfree( header );
	}
	do
	{
		refresh = FALSE;

		if ( (stream =  http_request( host, NULL, page, 80, ip, NULL, FALSE )) == NULL )
		{
			ERROR( "Can't get index page" );
			return -1;
		}
		if ( (fd = xfileno( stream )) == -1 )
		{
			ERROR( "Can't get file descriptor" );
      	return -1;
		}
		while ( net_fgets( buf, sizeof(buf), stream ) != NULL ) 
		{
			if ( (newurl = get_refresh( buf )) )
			{
				char * vhost = url_get_host( newurl );
				char * vpage = url_get_link( newurl );

				port 		= 	url_get_port( newurl );
				refresh 	= 	TRUE;

				if ( vhost ) 
				{
					xfree( host );
					host = xstrdup( vhost ); 
					xfree( vhost );
				}
				if ( vpage ) 
				{ 
					xfree( page ); 
					page = xstrdup( vpage ); 
					xfree( vpage );
				}
				if ( (ip = resolve( host )) == -1 )
				{
					ERROR( "Can't resolve refresh host" );
					return -1;
				}
				xfree( newurl );
			}
			parse_links( buf, host, ip_to_ascii( ip ), dir );
			parse_form( stream, buf, host, ip_to_ascii( ip ), dir );
		}
		XCLOSE( fd, stream );
	}
	while( refresh == TRUE && count++ < MAXRECURSION );

	xfree( page );

	do
	{
		allvisited = TRUE;

		for( htnode = hash_get_next( HTlinks, HTlinks[0] ); htnode ; htnode = hash_get_next( HTlinks, htnode ) )
		{
			struct linkz * l = (struct linkz *)(htnode->value);
			if ( !(l->flags & LVISITED) )
			{
				allvisited = FALSE;
				l->flags |= LVISITED;
				if ( (stream = http_request( host, NULL, htnode->key, 80, ip, &status, FALSE )) ) ;
				{
					if ( (fd = xfileno( stream )) == -1 )
      			{
						ERROR( "Can't get file descriptor" );
						return -1;
      			}
					if ( status == NOTFOUND )
					{
						XCLOSE( fd, stream );
						continue;
					}
      			while ( net_fgets( buf, sizeof(buf), stream ) != NULL )
      			{
						xfree( dir );
						dir = GETDIR( htnode->key );
         			parse_links( buf, host, ip_to_ascii( ip ), dir );
         			parse_form( stream, buf, host, ip_to_ascii( ip ), dir );
      			}
					XCLOSE( fd, stream );
   			}
			}
			hack_forms( host, 80 );
			hack_link( host, 80, htnode );
		}
	}
	while( allvisited == FALSE );

	hack_forms( host, 80 );
	hack_links( host, 80 );

	fprintf( stdout, DEF "\n\n -- End of Scan\n");
	return 1;
}
